@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-# Copyright (C) 2011-2013, Antonio Valentino <antonio.valentino@tiscali.it>
+# Copyright (C) 2011-2014, Antonio Valentino <antonio.valentino@tiscali.it>
 #
 # This file is part of PyEPR.
 #
@@ -24,14 +24,25 @@ import re
 import sys
 import numbers
 import operator
-import unittest
 import tempfile
 import functools
+from distutils.version import LooseVersion
+
+try:
+    import resource
+except ImportError:
+    resource = None
+
+try:
+    from unittest import skipIf
+except ImportError:
+    import unittest2 as unittest
+else:
+    del skipIf
+    import unittest
 
 import numpy as np
-
-TESTDIR = os.path.abspath(os.path.dirname(__file__))
-sys.path.insert(0, os.path.join(TESTDIR, os.pardir))
+import numpy.testing as npt
 
 import epr
 
@@ -52,6 +63,19 @@ EPR_TO_NUMPY_TYPE = {
 }
 
 
+def has_epr_c_bug_pyepr009():
+    v = LooseVersion(epr.EPR_C_API_VERSION)
+    if 'pyepr' in v.version:
+        return v < LooseVersion('2.3dev_pyepr082')
+    else:
+        return v <= LooseVersion('2.3')
+
+
+EPR_C_BUG_PYEPR009 = has_epr_c_bug_pyepr009()
+EPR_C_BUG_BCEPR002 = EPR_C_BUG_PYEPR009
+
+
+TESTDIR = os.path.abspath(os.path.dirname(__file__))
 TEST_PRODUCT = 'MER_LRC_2PTGMV20000620_104318_00000104X000_00000_00000_0001.N1'
 
 
@@ -573,7 +597,6 @@ class TestDataset(unittest.TestCase):
         created_record = self.dataset.create_record()
         read_record = self.dataset.read_record(0, created_record)
         self.assertTrue(created_record is read_record)
-        # @TODO: check contents
 
     def test_read_record_passed_invalid(self):
         self.assertRaises(TypeError, self.dataset.read_record, 0, 0)
@@ -640,8 +663,7 @@ class TestDatasetOnClosedProduct(unittest.TestCase):
         self.assertTrue(isinstance(self.dataset.product, epr.Product))
 
     def test_description_property(self):
-        # @TODO: check
-        self.assertEqual(self.dataset.description, '')
+        self.assertRaises(ValueError, getattr, self.dataset, 'description')
 
     def test_get_name(self):
         self.assertRaises(ValueError, self.dataset.get_name)
@@ -801,11 +823,11 @@ class TestBand(unittest.TestCase):
         self.assertRaises((ValueError, OverflowError),
                           self.band.create_compatible_raster,
                           width, -1)
-        # @TODO: check
-        #self.assertRaises(ValueError, self.band.create_compatible_raster,
-        #                  self.product.get_scene_width() + 10, height)
-        #self.assertRaises(ValueError, self.band.create_compatible_raster,
-        #                  width, self.product.get_scene_height() + 10)
+
+        self.assertRaises(ValueError, self.band.create_compatible_raster,
+                          self.product.get_scene_width() + 10, height)
+        self.assertRaises(ValueError, self.band.create_compatible_raster,
+                          width, self.product.get_scene_height() + 10)
 
     def test_create_compatible_raster_with_step(self):
         src_width = self.product.get_scene_width()
@@ -836,60 +858,103 @@ class TestBand(unittest.TestCase):
                           self.band.create_compatible_raster,
                           width, height, -2, -1)
 
-        # @TODO: check
-        #self.assertRaises((ValueError, OverflowError),
-        #                  self.band.create_compatible_raster,
-        #                  width, height, width + 10, 2)
-        #self.assertRaises((ValueError, OverflowError),
-        #                  self.band.create_compatible_raster,
-        #                  width, height, 2, height + 10)
-        #self.assertRaises((ValueError, OverflowError),
-        #                  self.band.create_compatible_raster,
-        #                  width, height, width + 10, height + 10)
+        self.assertRaises((ValueError, ValueError),
+                          self.band.create_compatible_raster,
+                          width, height, width + 10, 2)
+        self.assertRaises((ValueError, ValueError),
+                          self.band.create_compatible_raster,
+                          width, height, 2, height + 10)
+        self.assertRaises((ValueError, ValueError),
+                          self.band.create_compatible_raster,
+                          width, height, width + 10, height + 10)
+
+    def test_create_compatible_raster_with_default_size(self):
+        src_width = self.product.get_scene_width()
+        src_height = self.product.get_scene_height()
+        raster = self.band.create_compatible_raster()
+        self.assertTrue(isinstance(raster, epr.Raster))
+        self.assertEqual(raster.get_width(), src_width)
+        self.assertEqual(raster.get_height(), src_height)
+        # @NOTE: data type on disk is epr.E_TID_USHORT
+        self.assertEqual(raster.data_type, epr.E_TID_FLOAT)
 
     def test_read_raster(self):
-        raster = self.band.create_compatible_raster(self.WIDTH, self.HEIGHT)
+        rasterin = self.band.create_compatible_raster(self.WIDTH, self.HEIGHT)
 
-        self.band.read_raster(self.XOFFSET, self.YOFFSET, raster)
+        raster = self.band.read_raster(self.XOFFSET, self.YOFFSET, rasterin)
 
         self.assertTrue(isinstance(raster, epr.Raster))
+        self.assertTrue(raster is rasterin)
         self.assertEqual(raster.get_width(), self.WIDTH)
         self.assertEqual(raster.get_height(), self.HEIGHT)
         # @NOTE: data type on disk is epr.E_TID_USHORT
         self.assertEqual(raster.data_type, epr.E_TID_FLOAT)
 
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(raster.get_pixel(0, 0), self.TEST_DATA[0, 0])
+        npt.assert_allclose(raster.get_pixel(w - 1, h - 1),
+                            self.TEST_DATA[h - 1, w - 1])
+
+    def test_read_raster_none(self):
+        raster = self.band.read_raster()
+
+        self.assertTrue(isinstance(raster, epr.Raster))
+        self.assertEqual(raster.get_width(), self.product.get_scene_width())
+        self.assertEqual(raster.get_height(), self.product.get_scene_height())
+        # @NOTE: data type on disk is epr.E_TID_USHORT
+        self.assertEqual(raster.data_type, epr.E_TID_FLOAT)
+
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            raster.get_pixel(self.XOFFSET, self.YOFFSET), self.TEST_DATA[0, 0])
+        npt.assert_allclose(
+            raster.get_pixel(self.XOFFSET + w - 1, self.YOFFSET + h - 1),
+            self.TEST_DATA[h - 1, w - 1])
+
     def test_read_raster_default_offset(self):
-        height, width = self.TEST_DATA.shape
+        height = self.HEIGHT
+        width = self.WIDTH
 
         raster1 = self.band.create_compatible_raster(width, height)
         raster2 = self.band.create_compatible_raster(width, height)
 
-        self.band.read_raster(0, 0, raster1)
-        self.band.read_raster(raster=raster2)
+        r1 = self.band.read_raster(0, 0, raster1)
+        r2 = self.band.read_raster(raster=raster2)
 
+        self.assertTrue(r1 is raster1)
+        self.assertTrue(r2 is raster2)
         self.assertEqual(raster1.get_pixel(0, 0), raster2.get_pixel(0, 0))
         self.assertEqual(raster1.get_pixel(width - 1, height - 1),
                          raster2.get_pixel(width - 1, height - 1))
 
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            raster1.get_pixel(self.XOFFSET, self.YOFFSET),
+            self.TEST_DATA[0, 0])
+        npt.assert_allclose(
+            raster1.get_pixel(self.XOFFSET + w - 1, self.YOFFSET + h - 1),
+            self.TEST_DATA[h - 1, w - 1])
+
     def test_read_raster_with_invalid_raster(self):
         self.assertRaises(TypeError, self.band.read_raster, 0, 0, 0)
 
     def test_read_raster_with_invalid_offset(self):
         raster = self.band.create_compatible_raster(self.WIDTH, self.HEIGHT)
 
-        # @TODO: check
         self.assertRaises(ValueError, self.band.read_raster, -1, 0, raster)
         self.assertRaises(ValueError, self.band.read_raster, 0, -1, raster)
         self.assertRaises(ValueError, self.band.read_raster, -1, -1, raster)
-        # @TODO: check
-        #self.assertRaises(ValueError, self.band.read_raster,
-        #                  width + 10, 0, raster)
-        #self.assertRaises(ValueError, self.band.read_raster,
-        #                  0, height + 10, raster)
-        #self.assertRaises(ValueError, self.band.read_raster,
-        #                  width + 10, height + 10, raster)
 
-    def test_read_as_array(self):
+        invalid_xoffset = self.WIDTH
+        invalid_yoffset = self.HEIGHT
+        self.assertRaises(ValueError, self.band.read_raster,
+                          invalid_xoffset, 0, raster)
+        self.assertRaises(ValueError, self.band.read_raster,
+                          0, invalid_yoffset, raster)
+        self.assertRaises(ValueError, self.band.read_raster,
+                          invalid_xoffset, invalid_yoffset, raster)
+
+    def test_read_as_array_ref(self):
         data = self.band.read_as_array(self.WIDTH, self.HEIGHT,
                                        self.XOFFSET, self.YOFFSET)
 
@@ -898,29 +963,178 @@ class TestBand(unittest.TestCase):
         self.assertEqual(data.dtype, self.DATA_TYPE)
 
         h, w = self.TEST_DATA.shape
-        self.assertTrue(np.allclose(data[:h, :w], self.TEST_DATA))
+        npt.assert_allclose(data[:h, :w], self.TEST_DATA)
+
+    def test_read_as_array_cross(self):
+        data = self.band.read_as_array()
+        box = self.band.read_as_array(self.WIDTH, self.HEIGHT,
+                                      self.XOFFSET, self.YOFFSET)
+
+        npt.assert_array_equal(
+            data[self.YOFFSET:self.YOFFSET + self.HEIGHT,
+                 self.XOFFSET:self.XOFFSET + self.WIDTH],
+            box)
+
+    def test_read_as_array_default(self):
+        data = self.band.read_as_array()
+
+        self.assertTrue(isinstance(data, np.ndarray))
+        self.assertEqual(
+            data.shape,
+            (self.product.get_scene_height(), self.product.get_scene_width()))
+        self.assertEqual(data.dtype, self.DATA_TYPE)
+
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            data[self.YOFFSET:self.YOFFSET + h, self.XOFFSET:self.XOFFSET + w],
+            self.TEST_DATA)
 
-    # @TODO: check, it seems to be an upstream bug or a metter of data
-    #        mirroring
     # @SEEALSO: https://www.brockmann-consult.de/beam-jira/browse/EPR-2
-    #def test_read_as_array_with_step(self):
-    #    width = 400
-    #    height = 300
-    #
-    #    data = self.band.read_as_array(width, height,
-    #                                   self.XOFFSET, self.YOFFSET, 2, 2)
-    #
-    #    self.assertTrue(isinstance(data, np.ndarray))
-    #    self.assertEqual(data.shape, (height/2, width/2))
-    #    self.assertEqual(data.dtype, self.DATA_TYPE)
-    #
-    #    h, w = self.TEST_DATA.shape
-    #    self.assertTrue(
-    #               np.all(data[:h / 2, :w / 2] == self.TEST_DATA[::2, ::2]))
-    #    #self.assertTrue(
-    #    #           np.all(data[:h / 2, :w / 2] == self.TEST_DATA[::2, 1::2]))
+    @unittest.skipIf(EPR_C_BUG_BCEPR002, 'buggy EPR_C_API detected')
+    def test_read_as_array_with_step_2(self):
+        step = 2
+        band = self.band
 
-    # @TODO: more read_as_array testing
+        data = band.read_as_array()
+        box = band.read_as_array(self.WIDTH, self.HEIGHT,
+                                 self.XOFFSET, self.YOFFSET, step, step)
+
+        self.assertTrue(isinstance(box, np.ndarray))
+        self.assertEqual(
+            box.shape,
+            ((self.HEIGHT - 1) // step + 1, (self.WIDTH - 1) // step + 1))
+        self.assertEqual(box.dtype, self.DATA_TYPE)
+
+        npt.assert_allclose(
+            data[self.YOFFSET:self.YOFFSET + self.HEIGHT:step,
+                 self.XOFFSET:self.XOFFSET + self.WIDTH:step],
+            box)
+
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            box[:(h-1)//step+1, :(w-1)//step+1],
+            self.TEST_DATA[::step, ::step])
+
+    @unittest.skipIf(EPR_C_BUG_BCEPR002, 'buggy EPR_C_API detected')
+    def test_read_as_array_with_step_3(self):
+        step = 3
+        band = self.band
+
+        data = band.read_as_array()
+        box = band.read_as_array(self.WIDTH, self.HEIGHT,
+                                 self.XOFFSET, self.YOFFSET, step, step)
+
+        self.assertTrue(isinstance(box, np.ndarray))
+        self.assertEqual(
+            box.shape,
+            ((self.HEIGHT - 1) // step + 1, (self.WIDTH - 1) // step + 1))
+        self.assertEqual(box.dtype, self.DATA_TYPE)
+
+        npt.assert_allclose(
+            data[self.YOFFSET:self.YOFFSET + self.HEIGHT:step,
+                 self.XOFFSET:self.XOFFSET + self.WIDTH:step],
+            box)
+
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            box[:(h-1)//step+1, :(w-1)//step+1],
+            self.TEST_DATA[::step, ::step])
+
+    @unittest.skipIf(EPR_C_BUG_BCEPR002, 'buggy EPR_C_API detected')
+    def test_read_as_array_with_step_4(self):
+        step = 4
+        band = self.band
+
+        data = band.read_as_array()
+        box = band.read_as_array(self.WIDTH, self.HEIGHT,
+                                 self.XOFFSET, self.YOFFSET, step, step)
+
+        self.assertTrue(isinstance(box, np.ndarray))
+        self.assertEqual(
+            box.shape,
+            ((self.HEIGHT - 1) // step + 1, (self.WIDTH - 1) // step + 1))
+        self.assertEqual(box.dtype, self.DATA_TYPE)
+
+        npt.assert_allclose(
+            data[self.YOFFSET:self.YOFFSET + self.HEIGHT:step,
+                 self.XOFFSET:self.XOFFSET + self.WIDTH:step],
+            box)
+
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            box[:(h-1)//step+1, :(w-1)//step+1],
+            self.TEST_DATA[::step, ::step])
+
+    @unittest.skipIf(EPR_C_BUG_BCEPR002, 'buggy EPR_C_API detected')
+    def test_read_as_array_with_step_5(self):
+        step = 5
+        band = self.band
+
+        data = band.read_as_array()
+        box = band.read_as_array(self.WIDTH, self.HEIGHT,
+                                 self.XOFFSET, self.YOFFSET, step, step)
+
+        self.assertTrue(isinstance(box, np.ndarray))
+        self.assertEqual(
+            box.shape,
+            ((self.HEIGHT - 1) // step + 1, (self.WIDTH - 1) // step + 1))
+        self.assertEqual(box.dtype, self.DATA_TYPE)
+
+        npt.assert_allclose(
+            data[self.YOFFSET:self.YOFFSET + self.HEIGHT:step,
+                 self.XOFFSET:self.XOFFSET + self.WIDTH:step],
+            box)
+
+        h, w = self.TEST_DATA.shape
+        npt.assert_allclose(
+            box[:(h-1)//step+1, :(w-1)//step+1],
+            self.TEST_DATA[::step, ::step])
+
+
+class TestAnnotationBand(TestBand):
+    BAND_NAME = 'sun_zenith'
+    BAND_DESCTIPTION = 'Sun zenith angle'
+    SCALING_FACTOR = 9.999999974752427e-07
+    SCALING_OFFSET = 0.0
+    UNIT = 'deg'
+    TEST_DATA = np.asarray([
+        [33.11141205, 33.07904434, 33.04668045, 33.01435089, 32.98202515,
+         32.94969559, 32.91736603, 32.88507462, 32.85278320, 32.82049179],
+        [33.08905792, 33.05667114, 33.02428818, 32.99193954, 32.95959473,
+         32.92724609, 32.89489746, 32.86258316, 32.83027649, 32.79796219],
+        [33.06670380, 33.03429794, 33.00189590, 32.96953201, 32.93716431,
+         32.90479279, 32.87242889, 32.84009933, 32.80776978, 32.77543640],
+        [33.04434967, 33.01192474, 32.97950363, 32.94711685, 32.91473389,
+         32.88234329, 32.84995651, 32.81761169, 32.78525925, 32.75291061],
+        [33.02199554, 32.98955536, 32.95711136, 32.92470551, 32.89229965,
+         32.85989380, 32.82748795, 32.79512024, 32.76274872, 32.73038101],
+        [32.99978256, 32.96732330, 32.93486023, 32.90243149, 32.87001038,
+         32.83758163, 32.80516052, 32.77277374, 32.74037933, 32.70799255],
+        [32.97756958, 32.94509125, 32.91260529, 32.88016129, 32.84771729,
+         32.81527328, 32.78282928, 32.75042343, 32.71801376, 32.68560410],
+        [32.95535278, 32.92285538, 32.89035416, 32.85789108, 32.82542419,
+         32.79296494, 32.76049805, 32.72807693, 32.69564438, 32.66321945],
+        [32.93313980, 32.90061951, 32.86810303, 32.83562088, 32.80313873,
+         32.77065277, 32.73817062, 32.70572281, 32.67327881, 32.64083099],
+        [32.91106796, 32.87852859, 32.84599304, 32.81349182, 32.78099060,
+         32.74848557, 32.71598434, 32.68351364, 32.65105438, 32.61858749],
+    ])
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_read_raster(self):
+        super(TestAnnotationBand, self).test_read_raster()
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_read_raster_default_offset(self):
+        super(TestAnnotationBand, self).test_read_raster_default_offset()
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_read_as_array_ref(self):
+        super(TestAnnotationBand, self).test_read_as_array_ref()
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_read_as_array_cross(self):
+        super(TestAnnotationBand, self).test_read_as_array_cross()
 
 
 class TestBandHighLevelAPI(unittest.TestCase):
@@ -996,7 +1210,8 @@ class TestBandOnClosedProduct(unittest.TestCase):
 
     def test_lines_mirrored_property(self):
         self.assertTrue(isinstance(self.band.lines_mirrored, bool))
-        self.assertEqual(self.band.lines_mirrored, False)
+        # @TODO: check
+        #self.assertEqual(self.band.lines_mirrored, False)
 
     # END: check
 
@@ -1158,7 +1373,7 @@ class TestRaster(unittest.TestCase):
         self.assertEqual(data.dtype, EPR_TO_NUMPY_TYPE[self.raster.data_type])
 
         ny, nx = self.TEST_DATA.shape
-        self.assertTrue(np.allclose(data[:ny, :nx], self.TEST_DATA))
+        npt.assert_allclose(data[:ny, :nx], self.TEST_DATA)
 
     def test_data_property_two_times(self):
         data1 = self.raster.data
@@ -1186,7 +1401,7 @@ class TestRaster(unittest.TestCase):
         self.assertTrue(isinstance(data, np.ndarray))
 
         ny, nx = self.TEST_DATA.shape
-        self.assertTrue(np.allclose(data[:ny, :nx], self.TEST_DATA))
+        npt.assert_allclose(data[:ny, :nx], self.TEST_DATA)
 
 
 class TestRasterRead(TestRaster):
@@ -1215,6 +1430,26 @@ class TestRasterRead(TestRaster):
         self.assertTrue(np.all(data1 == data2))
 
 
+class TestAnnotatedRasterRead(TestRasterRead):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    BAND_NAME = TestAnnotationBand.BAND_NAME
+    RASTER_XOFFSET = TestAnnotationBand.XOFFSET
+    RASTER_YOFFSET = TestAnnotationBand.YOFFSET
+    TEST_DATA = TestAnnotationBand.TEST_DATA
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_get_pixel(self):
+        super(TestAnnotatedRasterRead, self).test_get_pixel()
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_data_property(self):
+        super(TestAnnotatedRasterRead, self).test_data_property()
+
+    @unittest.skipIf(EPR_C_BUG_PYEPR009, 'buggy EPR_C_API detected')
+    def test_data_property_raster_scope(self):
+        super(TestAnnotatedRasterRead, self).test_data_property_raster_scope()
+
+
 class TestRasterHighLevelAPI(unittest.TestCase):
     RASTER_WIDTH = 400
     RASTER_HEIGHT = 300
@@ -1530,8 +1765,7 @@ class TestField(unittest.TestCase):
         self.assertTrue(isinstance(vect, np.ndarray))
         self.assertEqual(vect.shape, (self.field.get_num_elems(),))
         self.assertEqual(vect.dtype, np.int8)
-        self.assertTrue(np.allclose(vect[:len(self.FIELD_VALUES)],
-                                    self.FIELD_VALUES))
+        npt.assert_allclose(vect[:len(self.FIELD_VALUES)], self.FIELD_VALUES)
 
 
 class TestFieldWithMiltipleElems(TestField):
@@ -1973,6 +2207,24 @@ class TestLibVersion(unittest.TestCase):
         self.assertTrue(isinstance(epr.EPR_C_API_VERSION, str))
 
 
+@unittest.skipIf(resource is None, '"resource" module not available')
+class TestMemoryLeaks(unittest.TestCase):
+    # See gh-10 (https://github.com/avalentino/pyepr/issues/10)
+
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+
+    def test_memory_leacks_on_read_as_array(self):
+        N = 10
+
+        for n in range(N):
+            with epr.open(self.PRODUCT_FILE) as p:
+                p.get_band('l2_flags').read_as_array()
+            if n <= 1:
+                m1 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
+            m2 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
+        self.assertLessEqual(m2 - m1, 8)
+
+
 if __name__ == '__main__':
     print('PyEPR: %s' % epr.__version__)
     print('EPR API: %s' % epr.EPR_C_API_VERSION)
